/* playcd.c  Callback Functions */
/*                            AppBuilder Photon Code Lib */
/*                                         Version 1. 0  */

/* Standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

/* Toolkit headers */
#include <Ph.h>
#include <Pt.h>
#include <Ap.h>

/* Local headers */
#define DEFINE_GLOBALS
#include "playcd.h"
#include "abimport.h"
#include "proto.h"


/******** POP UP ERROR DIALOG ********/
/* if errcode is 0, error message is displayed, but application is not terminated */
/* if errcode is not 0, application is terminated with errcode */
/* if errcode is 99, application is terminated with success */
/* if application is terminated, resources are released first */
void
error_popup (BYTE *errmsg, int errcode)
{
	if (errmsg != NULL){
		PtAskQuestion (
			ABW_base,
			NULL,
			errmsg,
			"lu18",
			"    OK    ",
			NULL,
			NULL,
			1
		);
	}
	if (errcode){
		/* terminate all operations in progress */
		stop (NULL, NULL, NULL);
		
		/* Release Resources */
		if (fd_cd != -1){
			close (fd_cd);
		}
		if (timer_id != -1){
			timer_delete (timer_id);
		}
		if (timer_proxy != -1){
			qnx_proxy_detach (timer_proxy);
		}
		if (errcode == 99){
			errcode = 0;
		}
		exit (errcode);
	}
} /* end error_popup() */


/******** APPLICATION INITIALIZATION ********/
int
init (int argc, char *argv[])
{
	/* eliminate 'unreferenced' warnings */
	argc = argc, argv = argv;
	
	{ /* Read Configuration */
		FILE *fp_config;
		char fspec_config[81];
		
		sprintf (fspec_config, cfg_file, getnid ());
		
		if ((fp_config = fopen (fspec_config, "r"))){
			char config_data[81];
			
			while (fgets (config_data, sizeof (config_data), fp_config) == config_data){
				/* Get I/O Port Base */
				if ( !memcmp (config_data, "Port=", 5)){
					WORD config_base;
					WORD *port_bases_pt;
					
					config_base = atoh (config_data + 5) + 4;
					port_bases_pt = port_bases;
					
					while (*port_bases_pt){
						if (config_base == *port_bases_pt){
							mixer_base = config_base;
							break;
						}
						++port_bases_pt; /* try next */
					}
					/* If port not matched, left at default */
				}
				/* Get CD-ROM Drive */
				if ( !memcmp (config_data, "CD=", 3)){
					BYTE *cdpt;
					
					cdpt = config_data + 3;
					while (*cdpt){
						if (*cdpt <= ' ' || *cdpt >= 0x7F){
							*cdpt = 0;
							break;
						}
						++cdpt;
					}
					strncpy (devcd, config_data + 3, 26);
					devcd[25] = 0;
				}
				if ( !memcmp (config_data, "Card=SoundBlaster16", 19)){
					card_type = SB16;
				}
				if ( !memcmp (config_data, "Card=SoundBlasterPro", 20)){
					card_type = SBP;
				}
			}
			fclose (fp_config);
		}
		/* Else No configuration file */
		/* Leave set at initial values */
	}
	
	/* Initialize Audio Card */
	audio_update ();
	
	playlist[0] = 0; /* no tracks selected yet */
	
	/* Create Proxy And Timer, Set Up Photon To Direct Messages To Callback */
	if ((timer_proxy = qnx_proxy_attach( 0, NULL, 0, -1 )) == 0){
		error_popup ("Failed to create a proxy", EXIT_FAILURE);
	}
	else {
		/* Create Timer To Trigger Proxy */
		struct sigevent timer_ev;
		
		timer_ev.sigev_signo = -timer_proxy;
		if ((timer_id = timer_create (CLOCK_REALTIME, &timer_ev)) == 0){
			error_popup ("Failed to create a timer", EXIT_FAILURE);
		}
		PtAppAddInput (NULL, timer_proxy, play_next_track, NULL );
	}
	
	return (Pt_CONTINUE);

} /* end init() */


/******** OPEN CD-ROM DEVICE ********/
/* Returns -1 if failed to open */
int
open_cd ()
{
	if (fd_cd == -1){
		fd_cd = open (devcd, O_RDONLY);
		if (fd_cd == -1){
			sprintf (errbuf, "Could not open '%s'\nWas driver started ?", devcd);
			error_popup (errbuf, 0);
		}
	}
	return (fd_cd);

} /* end open_cd() */
	
	
/******** PREPARE HARDWARE CONFIGURATION DIALOG ********/
int
config_setup (PtWidget_t *link_instance, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	/* eliminate 'unreferenced' warnings */
	link_instance = link_instance, apinfo = apinfo, cbinfo = cbinfo;
	
	/* IDENTIFY CURRENT I/O PORT */
	{
		WORD	sel_array[1];	/* list of selected items */
		WORD	*pt_port_base;	/* pointer to list of valid I/O Ports */
		
		/* Determine Current Setting */
		sel_array[0] = 1;
		pt_port_base = port_bases;
		
		while (*pt_port_base){
			if (*pt_port_base == mixer_base){
				break;
			}
			++sel_array[0];
			++pt_port_base;
		}
		
		if (*pt_port_base == 0){
			/* Invalid Base, Select Default */
			pt_port_base = port_bases;
			sel_array[0] = 1;
		}
		
		/* Set Default To Be Previous Value */
		new_mixer_base = *pt_port_base;
		
		PtSetArg (&args[0], Pt_ARG_SELECTION_INDEXES, sel_array, 1);
		PtSetResources (ABW_PtList_ioport, 1, args);
	}
	
	/* IDENTIFY CARD TYPE */
	/* Fix Invalid Card Type - Set To Default */
	if (card_type < SB16 && card_type > SBP){
		card_type = SB16;
	}
	
	/* Set Default To Be Previous Value */
	new_card_type = card_type;
	
	PtSetArg (&args[0], Pt_ARG_CBOX_SELECTION_ITEM, card_type, 0);
	PtSetResources (ABW_PtComboBox_card, 1, args);
	
	/* IDENTIFY CD-ROM BLOCK SPECIAL DEVICE */
	/* Fix Undefined Device - Set To Default */
	if (strlen (devcd) == 0){
		strcpy (devcd, DEFAULT_CD);
	}
	
	/* Set Default To Current Value */
	strcpy (new_devcd, devcd);
	
	PtSetArg (&args[0], Pt_ARG_TEXT_STRING, devcd, 0);
	PtSetResources (ABW_PtText_device, 1, args);

	return (Pt_CONTINUE);
} /* end config_setup() */


/******** SELECT NEW I/O PORT  ********/
int
change_ioport (PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	PtListCallback_t	*portinfo = cbinfo->cbdata;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo;
	
	if (cbinfo->reason_subtype == Pt_LIST_SELECTION_FINAL){

		/* Set new port base */
		/* Not actually updated until Done */
		new_mixer_base = port_bases[portinfo->item_pos - 1];
	}
	
	return (Pt_CONTINUE);
} /* end select_ioport() */


/******** SELECT NEW BLOCK SPECIAL DEVICE ********/
int
new_device( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;
	
	if (cbinfo->reason == Pt_CB_TEXT_CHANGED){
		/* Not Actually Updated Until Done */
		strcpy (new_devcd, ((PtTextCallback_t *)cbinfo->cbdata)->text);
	}

	return (Pt_CONTINUE);
} /* end new_device() */


/******** SELECT NEW AUDIO CARD TYPE ********/
int
new_card( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;
	
	if (cbinfo->reason == Pt_CB_SELECTION){
		/* Not Actually Updated Until Done */
		new_card_type = ((PtComboBoxListCallback_t *)cbinfo->cbdata)->item_pos;
	}

	return (Pt_CONTINUE);
} /* end new_card() */


/******** APPLY NEW CONFIGURATION ********/
int
config_apply (PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/* Set new port base and initialize Sound Blaster */
	mixer_base = new_mixer_base;
	card_type = new_card_type;
	do_audio_init = 1;
	audio_update ();
	
	/* Set New CD Device */
	if (strcmp (devcd, new_devcd)){
		/* Close Previous Device */
		if (fd_cd != -1){
			close (fd_cd);
			fd_cd = -1;
		}
		strcpy (devcd, new_devcd);
	}
	
	{ /* Save Configuration */
		FILE *fp_config;
		char fspec_config[81];
		
		sprintf (fspec_config, cfg_file, getnid ());
		
		if ((fp_config = fopen (fspec_config, "w")) == NULL){
			
			char errmsg[81];
			
			sprintf (errmsg, "Could not open %s to save configuration\n", fspec_config);
			
			/* Complain About Config File Inaccessable with POPUP DIALOG */
			PtAskQuestion (
				ABW_base,
				NULL,
				errmsg,
				"helv14",
				"&Ok",
				NULL,
				NULL,
				1
			);
			
		}
		else {
			switch (card_type){
			case SB16:
				fprintf (fp_config, "Card=SoundBlaster16\n");
				break;
			case SBP:
				fprintf (fp_config, "Card=SoundBlasterPro\n");
				break;
			}
			fprintf (fp_config, "Port=%04X\n", mixer_base - 4);
			fprintf (fp_config, "CD=%s\n", devcd);
			fclose (fp_config);
		}
	}

	return (Pt_CONTINUE);
} /* end config_apply() */


/******** UPDATE SOUND BLASTER HARDWARE ********/
void
audio_update ()
{
	static FLAG card_not_init = 1;
	
	if (card_not_init){
		card_not_init = 0;
		do_audio_init = 1;
	}
	else {
		/* Detect Card Status To Identify If Not Initialized */
		switch (card_type){
		case SB16:
			outp (mixer_base, SB16_CD_VOLUME_L);
			if (inp (mixer_base + 1) == 0){
				/* CD audio not enabled, audio card requires initialization */
				do_audio_init = 1;
			}
			break;
		case SBP:
			break;
		}
	}

	/* Initialization Of Audio Card */
	if (do_audio_init){
		switch (card_type){
		case SB16:
			/* Turn off volume of undesired channels */
			set_mixer (SB16_VOICE_VOLUME_L, 0);
			set_mixer (SB16_VOICE_VOLUME_R, 0);
			set_mixer (SB16_MIDI_VOLUME_L, 0);
			set_mixer (SB16_MIDI_VOLUME_R, 0);
			set_mixer (SB16_LINE_VOLUME_L, 0);
			set_mixer (SB16_LINE_VOLUME_R, 0);
			set_mixer (SB16_MIC_VOLUME, 0);
			set_mixer (SB16_PC_SPEAKER_VOLUME, 0);
			
			/* Set gain of amplifiers */
			set_mixer (SB16_INPUT_GAIN_LEFT, 0x00);
			set_mixer (SB16_INPUT_GAIN_RIGHT, 0x00);
			set_mixer (SB16_OUTPUT_GAIN_LEFT, 0x00);
			set_mixer (SB16_OUTPUT_GAIN_RIGHT, 0x00);
			set_mixer (SB16_AUTO_GAIN_CONTROL, 0);
			
			/* Select desired audio paths (CD audio only) */
			set_mixer (SB16_INPUT_MIXER_LEFT, 0x00);
			set_mixer (SB16_INPUT_MIXER_RIGHT, 0x00);
			set_mixer (SB16_OUTPUT_MIXER, 0x06);
			
			set_mixer (SB16_CD_VOLUME_L, 0xF8);
			set_mixer (SB16_CD_VOLUME_R, 0xF8);
	
			/* Cause Update Of Registers */
			volume = 0x98;
			do_volume = 1;
			treble = 0x60;
			do_treble = 1;
			bass = 0xA0;
			do_bass = 1;
			break;
		
		case SBP:
			/* Turn off volume of undesired channels */
			set_mixer (SBP_VOL_UNKNOWN1, 11);
			set_mixer (SBP_VOL_UNKNOWN2, 11);
			
			/* Select desired audio paths (CD audio only) */
			set_mixer (SBP_INPUT_MIXER, 0x1B);
			
			set_mixer (SBP_CD_VOLUME, 0xFF);
	
			/* Cause Update Of Registers */
			volume = 0x99;
			do_volume = 1;
			break;
		}

		do_audio_init = 0; /* done */
	} /* end if (do_audio_init) */
			
	/* Set Treble Controls */
	if (do_treble){
		switch (card_type){
		case SB16:
			set_mixer (SB16_TREBLE_LEFT, treble);
			set_mixer (SB16_TREBLE_RIGHT, treble);
			break;
		case SBP: /* no hardware support */
			break;
		}
		do_treble = 0;
	}
	
	/* Set Bass Controls */
	if (do_bass){
		switch (card_type){
		case SB16:
			set_mixer (SB16_BASS_LEFT, bass);
			set_mixer (SB16_BASS_RIGHT, bass);
			break;
		case SBP: /* no hardware support */
			break;
		}
		do_bass = 0;
	}
	
	/* Set Volume Controls */
	if (do_volume){
		switch (card_type){
		case SB16:
			set_mixer (SB16_MASTER_VOLUME_L, volume);
			set_mixer (SB16_MASTER_VOLUME_R, volume);
			break;
		case SBP:
			set_mixer (SBP_MASTER_VOLUME, volume);
			break;
		}
		do_volume = 0;
	}
} /* end audio_update() */

	
/******** OUTPUT VALUE TO SOUND BLASTER MIXER REGISTER ********/
void 
set_mixer (BYTE index, BYTE value)
{
	switch (card_type){
	case SB16:
	case SBP:
		outp (mixer_base, index);
		outp (mixer_base + 1, value);
		break;
	}

} /* end set_mixer() */


/******** CHANGE VOLUME SETTING ********/
int
change_volume (PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	PtSliderCallback_t	*slinfo = cbinfo->cbdata;
	int slider_val;
	
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo;
	
	slider_val = slinfo->position;
	
	if (slider_val > 15 || slider_val < 0){
		/* Slider value out of range 0-15 */
		return (Pt_CONTINUE);
	}
	
	switch (card_type){
	case SB16:
		volume = mixer_sb16[slider_val];
		break;
	case SBP:
		volume = mixer_sbp[slider_val];
		break;
	}
	do_volume = 1;
	
	audio_update ();

	return (Pt_CONTINUE);
} /* end change_volume() */


/******** MODIFY BASS SETTING ********/
int
change_bass (PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	PtSliderCallback_t	*slinfo = cbinfo->cbdata;
	int slider_val;
	
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo;
	
	slider_val = slinfo->position;
	
	if (slider_val > 15 || slider_val < 0){
		/* Slider value out of range 0-15 */
		return (Pt_CONTINUE);
	}
	
	bass = mixer_sb16[slider_val];
	do_bass = 1;
	
	audio_update ();

	return (Pt_CONTINUE);
} /* end change_bass() */


/******** MODIFY TREBLE SETTING ********/
int
change_treble (PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	PtSliderCallback_t	*slinfo = cbinfo->cbdata;
	int slider_val;
	
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo;
	
	slider_val = slinfo->position;
	
	if (slider_val > 15 || slider_val < 0){
		/* Slider value out of range 0-15 */
		return (Pt_CONTINUE);
	}
	
	treble = mixer_sb16[slider_val];
	do_treble = 1;
	
	audio_update ();

	return (Pt_CONTINUE);
} /* End change_treble() */


/******** PLAY THE TRACKS ********/
int
play (PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;
	
	if (do_resume){
		paused (NULL, NULL, NULL);
		return (Pt_CONTINUE);
	}

	if (do_load){
		/* CD Not Yet Loaded */
		load (NULL, NULL, NULL);
	}
		
	if ( !do_load){
		/* CD Loaded OK */
		if (playlist[0] == 0){
			/* No Tracks Selected Before Play */
			/* Add All Tracks To Playlist */
			BYTE *playlistpt;
			int tksel;
			
			playlistpt = playlist;
			
			for (tksel = 1; tksel <= total_tracks; tksel++){
				*playlistpt++ = tksel;
			}
			*playlistpt = 0;
		}
		
		/* Play First Track */
		play_next_track (NULL, 0, NULL, 0);
	}
		
	return (Pt_CONTINUE);
} /* end play() */


/******** PAUSE PLAYING CD ********/
int
paused (PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;
	
	/* Show Button As Not Paused */
	PtSetArg (&args[0], Pt_ARG_FILL_COLOR, FILL_COLOUR, 0);
	PtSetResources (ABW_PtButton_pause, 1, args);

	if (do_resume){
		if (do_resume){
			/* Start CD-ROM Again */
			if (open_cd () != -1){
				if (qnx_ioctl (fd_cd, CDROMRESUME, NULL, 0, NULL, 0) != -1){
					/* Start Timer With What Remains */
					timer_settime (timer_id, 0, &playtime, NULL);
				}
			}
			do_resume = 0;
		}
	}
	else if (playing_track) {
		/* Get Outstanding Time For Current Audio Track */
		if (timer_gettime (timer_id, &playtime) == 0){
			if ((playtime.it_value.tv_sec + playtime.it_value.tv_nsec) != 0L){

				/* Time Remaining, So Play In Progress, Pause CD */
				if (open_cd () != -1){
					if (qnx_ioctl (fd_cd, CDROMPAUSE, NULL, 0, NULL, 0) != -1){
						struct itimerspec stoptime;
						
						/* Stop Timer To Avoid Callback Being Triggered */
						stoptime.it_value.tv_sec = 0;
						stoptime.it_value.tv_nsec = 0;
						stoptime.it_interval.tv_sec = 0;
						stoptime.it_interval.tv_nsec = 0;
						timer_settime (timer_id, 0, &stoptime, NULL);
						
						do_resume = 1;

						/* Show Button As Paused */
						PtSetArg (&args[0], Pt_ARG_FILL_COLOR, ACTIVE_COLOUR, 0);
						PtSetResources (ABW_PtButton_pause, 1, args);
					}
				}
			}
		}
	}

	return (Pt_CONTINUE);

} /* end paused() */


/******** STOP PLAYING CD ********/
int
stop (PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;
	
	if (do_resume){
		do_resume = 0;
		/* Show Button As Not Paused */
		PtSetArg (&args[0], Pt_ARG_FILL_COLOR, FILL_COLOUR, 0);
		PtSetResources (ABW_PtButton_pause, 1, args);
	}
	
	playlist[0] = 0; /* no tracks selected to play */
	play_next_track (NULL, 0, NULL, 0);
	
	return (Pt_CONTINUE);
} /* end stop() */


/******** LOAD TRACK LIST FROM CD ********/
int
load (PtWidget_t *link_instance, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	int tkno; /* loop counter for track number */
	int errcode; /* -1 is failure, 0 is ok */
	
	/* eliminate 'unreferenced' warnings */
	link_instance = link_instance, apinfo = apinfo, cbinfo = cbinfo;
	
	errcode = 0;
	
	if (do_resume){
		do_resume = 0;
		/* Show Button As Not Paused */
		PtSetArg (&args[0], Pt_ARG_FILL_COLOR, FILL_COLOUR, 0);
		PtSetResources (ABW_PtButton_pause, 1, args);
	}
	
	/* Send Load Command To Device */
	trklist[0] = " Reading Disk ";
	PtSetArg (&args[0], Pt_ARG_ITEMS, trklist, 1);
	PtSetResources (ABW_PtList_tracks, 1, args);
	PtFlush ();

	/* LOAD CD-ROM */
	if (open_cd () != -1){
		/* Cannot test for error here since some drivers do not support load/eject */
		qnx_ioctl (fd_cd, CDROMLOAD, NULL, 0, NULL, 0);
	}
	else {
		errcode = -1;
	}
	
	if ( !errcode){
		/* Wait For The Driver To Initialize */
		int attempt;
		
		memset ( &tochdr, 0, sizeof (tochdr));
		attempt = 20;
		errcode = -1;
		
		while (attempt--){
			if (open_cd () != -1){
				if (qnx_ioctl (fd_cd, CDROMREADTOCHDR, NULL, 0, &tochdr, sizeof (tochdr)) == 0){
					/* Read Header Succeeded */
					errcode = 0;
					break;
				}
			}
			sleep (1); /* wait for driver to initialize */
		}
		
		if (errcode){
			error_popup ("Failed CD Read Header", 0);
		}
	}
		
	/* Load The Track Data */
	if ( !errcode){
		total_tracks = 0;
		memset (tkent, 0, sizeof (tkent));
		tkentpt = tkent;
		
		/* Limit Number Of Tracks To What Is Supported */
		/* Leave One For The Leadout Track */
		if ((tochdr.cdth_trk1 - tochdr.cdth_trk0 + 2) >= MAXTRKS){
			tochdr.cdth_trk1 = MAXTRKS + tochdr.cdth_trk0 - 2;
		}
	
		{ /* Read Data For Each Track */
			long last_lba;
			
			last_lba = 0;
			
			for (tkno = tochdr.cdth_trk0; tkno <= (tochdr.cdth_trk1 + 1); tkno++){
				/* Select Track Number To Read */
				if (tkno > tochdr.cdth_trk1){
					tkentpt->cdte_track = CDROM_LEADOUT;
				}
				else {
					tkentpt->cdte_track = tkno;
				}
				tkentpt->cdte_format = CDROM_LBA;
				if (open_cd () != -1){
					if (qnx_ioctl (fd_cd, CDROMREADTOCENTRY, tkentpt, sizeof (struct cdrom_tocentry), tkentpt, sizeof (struct cdrom_tocentry)) < 0){
						sprintf (errbuf, "Read Entry %d failed", tkno);
						error_popup (errbuf, 0);
						errcode = -1;
						break;
					}
				}
				else {
					errcode = -1;
				}
				
				/* Create The Text For The List Widget */
				if (tkno > tochdr.cdth_trk0){
					/* Duration is difference between this track and previous */
					long this_lba;
					
					/* Add 40 to round up at 40/75 of a second */
					this_lba = tkentpt->cdte_addr.lba - last_lba + 40;
					
					if ((tkentpt->cdte_ctrl & CDROM_DATA_TRACK) == 0){
						/* Audio Track */
						/* Add To The PtList Widget */
						trklist[total_tracks] = trklistdata + total_tracks * 12;
						sprintf (trklist[total_tracks], "%-3d%5ld:%02ld",
						  tkno - 1, this_lba / 4500, (this_lba / 75) % 60);
						++total_tracks;
					}
				}
				
				last_lba = tkentpt->cdte_addr.lba; /* save location of previous track */
				++tkentpt;
			} /* end for (tkno) */
		}
	}
	
	/* Update Track List Widget */
	if (errcode == -1){
		memset (tkent, 0, sizeof (tkent));
		trklist[0] = "       Error    ";
		trklist[1] = "  Reload Disk ";
		total_tracks = 2;
		do_load = 1;
	}
	else if (total_tracks == 0){
		memset (tkent, 0, sizeof (tkent));
		trklist[0] = "   Data Only";
		trklist[1] = " Load Another";
		total_tracks = 2;
		do_load = 1;
	}
	else {
		do_load = 0;
		playlist[0] = 0; /* no tracks selected to play */
	}
	
	PtSetArg (&args[0], Pt_ARG_ITEMS, trklist, total_tracks);
	PtSetResources (ABW_PtList_tracks, 1, args);
	
	if (errcode){
		total_tracks = 0;
	}
	
	/* Change Colour Of Scrollbar To Match Application Colours */
	if (total_tracks > 13){
		/* This requires a level of indirection of PtSetResources ! */
		/* That is, the Pt_ARG_LIST_SB_RES is a resource that is */
		/* an array of resources */
		PtArg_t sbargs[4];
		
		PtSetArg (&sbargs[0], Pt_ARG_FILL_COLOR, FILL_COLOUR, 0);
		PtSetArg (&sbargs[1], Pt_ARG_COLOR, FILL_COLOUR, 0);
		PtSetArg (&sbargs[2], Pt_ARG_TOP_BORDER_COLOR, HIGHLIGHT_COLOUR, 0);
		PtSetArg (&sbargs[3], Pt_ARG_BOT_BORDER_COLOR, SHADOW_COLOUR, 0);
		PtSetArg (&args[0], Pt_ARG_LIST_SB_RES, sbargs, 4);
		PtSetResources (ABW_PtList_tracks, 1, args);
	}

	return (Pt_CONTINUE);
} /* end load() */


/******** EJECT CD ********/
int
eject (PtWidget_t *link_instance, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	/* eliminate 'unreferenced' warnings */
	link_instance = link_instance, apinfo = apinfo, cbinfo = cbinfo;
	
	if (do_resume){
		do_resume = 0;
		/* Show Button As Not Paused */
		PtSetArg (&args[0], Pt_ARG_FILL_COLOR, FILL_COLOUR, 0);
		PtSetResources (ABW_PtButton_pause, 1, args);
	}
	
	/* Cancel Any Outstanding Play */
	playlist[0] = 0;
	play_next_track (NULL, 0, NULL, 0);
	
	/* Send Eject Command To CD-ROM */
	if (open_cd () != -1){
		/* Cannot test for error here since some drivers do not support load/eject */
		qnx_ioctl (fd_cd, CDROMEJECT, NULL, 0, NULL, 0);
	}
	
	/* Update Track List Widget */
	memset (tkent, 0, sizeof (tkent));
	trklist[0] = "  Load Disk  ";
	do_load = 1;
	
	PtSetArg (&args[0], Pt_ARG_ITEMS, trklist, 1);
	PtSetResources (ABW_PtList_tracks, 1, args);
	
	return (Pt_CONTINUE);
} /* end eject() */


/******** SELECT TRACKS TO PLAY ********/
int
select_track (PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	PtListCallback_t *playinfo = cbinfo->cbdata;
	
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;
	
	if (cbinfo->reason_subtype == Pt_LIST_SELECTION_FINAL){
		BYTE playtrk;
		BYTE *playlistpt;
		
		playlistpt = playlist;
		playtrk = (playinfo->item_pos);
		
		/* Check If Track Already Selected */
		while (*playlistpt){
			if (*playlistpt == playtrk){
				/* Matched */
				break;
			}
			++playlistpt;
		}
		
		if (*playlistpt == 0){
			/* New Selection */
			if ((playlistpt - playlist) < MAXTRKS){
				/* Add Selection */
				*playlistpt++ = playtrk;
				*playlistpt = 0;
			}
		}
		if (*playlistpt == playtrk){
			/* Delete Selection */
			while (*playlistpt){
				*playlistpt = *(playlistpt + 1);
				++playlistpt;
			}
		}
	}

	return (Pt_CONTINUE);
} /* end select_track() */


/******** PLAY NEXT TRACK IN PLAYLIST ********/
/* For subsequent tracks this is invoked by the clock proxy callback */
int
play_next_track (void *notused, pid_t pid, void *msg, size_t size)
{
	if (playlist[0] == 0){
		/* All Tracks Played Or Stop or Eject Clicked */
		
		/* Display No Tracks Selected */
		/* Permit Selections In Track List */
		PtSetArg (&args[0], Pt_ARG_SELECTION_INDEXES, NULL, 0);
		PtSetArg (&args[1], Pt_ARG_LIST_FLAGS, 0, Pt_LIST_NON_SELECT);
		PtSetArg (&args[2], Pt_ARG_CURSOR_TYPE, Ph_CURSOR_POINTER, 0);
		PtSetResources (ABW_PtList_tracks, 3, args);
		
		/* Set Play Button To Normal Colour */
		PtSetArg (&args[0], Pt_ARG_FILL_COLOR, FILL_COLOUR, 0);
		PtSetResources (ABW_PtButton_play, 1, args);
		
		{ /* Cancel Timer */
			playtime.it_value.tv_sec = 0;
			playtime.it_value.tv_nsec = 0;
			playtime.it_interval.tv_sec = 0;
			playtime.it_interval.tv_nsec = 0;
			timer_settime (timer_id, 0, &playtime, NULL);
		}
		
		/* Issue Stop Command To CD Drive */
		if (open_cd () != -1){
			qnx_ioctl (fd_cd, CDROMSTOP, NULL, 0, NULL, 0);
		}
		
		playing_track = 0;
		do_resume = 0;
		
		/* Reset Icon To Indicate No Current Track */
		sprintf (errbuf, "PLAYCD");
		PtSetArg (&args[0], Pt_ARG_WINDOW_TITLE, errbuf, 0);
		PtSetResources (ABW_base, 1, args);
	}
	else {
		playing_track = playlist[0];
		
		{ /* Delete track from beginning of list */
			BYTE *playlistpt;
			
			playlistpt = playlist;
			while (*playlistpt){
				*playlistpt = *(playlistpt + 1);
				++playlistpt;
			}
		}
		
		{ /* Show Current Track By Selecting It */
		  /* Prevent Selections In Track List */
			ushort_t trk_sel[1];
			
			trk_sel[0] = playing_track;
			PtSetArg (&args[0], Pt_ARG_SELECTION_INDEXES, trk_sel, 1);
			PtSetArg (&args[1], Pt_ARG_LIST_FLAGS, Pt_LIST_NON_SELECT, Pt_LIST_NON_SELECT);
			PtSetArg (&args[2], Pt_ARG_CURSOR_TYPE, Ph_CURSOR_NOINPUT, 0);
			PtSetResources (ABW_PtList_tracks, 3, args);
			
			{ /* Ensure Selected Track Is Within Displayed Portion */
				ushort_t *top_item;
				ushort_t new_top;
				
				PtSetArg (&args[0], Pt_ARG_TOP_ITEM_POS, &top_item, 0);
				PtGetResources (ABW_PtList_tracks, 1, args);
				new_top = *top_item;

				if (playing_track < new_top){
					PtSetArg (&args[0], Pt_ARG_TOP_ITEM_POS, playing_track, 0);
					PtSetResources (ABW_PtList_tracks, 1, args);
				}
				else if (playing_track > (new_top + 12)){
					while (playing_track > (new_top + 12)){
						++new_top;
					}
					PtSetArg (&args[0], Pt_ARG_TOP_ITEM_POS, new_top, 0);
					PtSetResources (ABW_PtList_tracks, 1, args);
				}
			}				
		}		

		/* Set Play Button To Indicate Play In Operation */
		PtSetArg (&args[0], Pt_ARG_FILL_COLOR, ACTIVE_COLOUR, 0);
		PtSetResources (ABW_PtButton_play, 1, args);
			
		{ /* Set Timer To Trigger This Callback When Track Finished */
		  /* Issue Playmsf Command To CD Drive */
			struct cdrom_msf msf;
			long start_lba;
			long finish_lba;
			
			tkentpt = &tkent[playing_track - 1];
			start_lba = tkentpt->cdte_addr.lba;
			finish_lba = (tkentpt + 1)->cdte_addr.lba;
			
			/* starting minute, second, frame */
			msf.cdmsf_min0 = LBA2MIN (start_lba);
			msf.cdmsf_sec0 = LBA2SEC (start_lba);
			msf.cdmsf_frame0 = LBA2FRAME (start_lba);
			/* ending minute, second, frame */
			msf.cdmsf_min1 = LBA2MIN (finish_lba);
			msf.cdmsf_sec1 = LBA2SEC (finish_lba);
			msf.cdmsf_frame1 = LBA2FRAME (finish_lba);
			
			/* Issue PLAYMSF Command To CD Drive */
			if (open_cd () != -1){
				if (qnx_ioctl (fd_cd, CDROMPLAYMSF, &msf, sizeof (msf), NULL, 0) != -1){
					/* Play Command Succeeded */
					/* Set Timer To Trigger When Track Completed */
					finish_lba = finish_lba - start_lba; /* duration */
					playtime.it_value.tv_sec = finish_lba / 75;
					playtime.it_value.tv_nsec = (finish_lba % 75) * 13333333;
					playtime.it_interval.tv_sec = 0;
					playtime.it_interval.tv_nsec = 0;
					timer_settime (timer_id, 0, &playtime, NULL);

					/* Set Icon To Indicate Current Track */
					sprintf (errbuf, "PLAYCD  [%d]", tkentpt->cdte_track);
					PtSetArg (&args[0], Pt_ARG_WINDOW_TITLE, errbuf, 0);
					PtSetResources (ABW_base, 1, args);
				}
				else {
					/* Failed - Probably Data Track Not Marked As Such */
					/* Update Track List */
					sprintf (trklist[playing_track - 1], "%-3d    data", tkentpt->cdte_track);
					PtSetArg (&args[0], Pt_ARG_ITEMS, trklist, total_tracks);
					PtSetResources (ABW_PtList_tracks, 1, args);
					
					/* Skip To Next Track */
					play_next_track (NULL, 0, NULL, 0);
				}
			}
		}
	}
		
	return (Pt_CONTINUE);
} /* end play_next_track() */


/******** END APPLICATION ********/
int
end (PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo)
{
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;
	
	if (((PhWindowEvent_t *)cbinfo->cbdata)->event_f == Ph_WM_CLOSE){
		/* This terminates program */
		error_popup (NULL, EXIT_SUCCESS);
	}
	return (Pt_CONTINUE);

} /* end end() */


