/* playcd.h  Definitions and Global Variables for playcd */

/* Check Message Displayed After Ejecting Disk */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <conio.h>
#include <fcntl.h>
#include <sys/cdrom.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/qioctl.h>
#include <sys/proxy.h>

/* Widget Custom Colours */
/* Item           Red    Green     Blue */
/* Background     190     190      255  */
/* Fill           155     155      255  */
/* Active         210     170      255  */
/* Highlight      210     210      255  */
/* Shadow         125     125      255  */

#define FILL_COLOUR			PgRGB(155,155,255)
#define ACTIVE_COLOUR		PgRGB(210,170,255)
#define HIGHLIGHT_COLOUR	PgRGB(210,210,255)
#define SHADOW_COLOUR		PgRGB(125,125,255)

#include "typedef.h"

#define MAXTRKS			99				/* maximum audio tracks supported */
#define TRK_LIST_CNT	13				/* number of tracks displayed in list */

/* defines for card_type */
#define NO_CARD			0				/* no audio card defined */
#define SB16			1				/* Sound Blaster 16, Vibra 16 */
#define SBP				2				/* Sound Blaster Pro */

/* SOUND BLASTER PRO DEFINITIONS */
/* These were determined experimentally */
#define SBP_MASTER_VOLUME	0x22		/* left nibble is left volume */
										/* right nibble is right volume */
										/* odd values permitted only */
										/* 1=low, F=high */
#define SBP_VOL_UNKNOWN1	0x26
#define SBP_VOL_UNKNOWN2	0x2E
#define SBP_CD_VOLUME		0x28		/* Values as MASTER */
#define SBP_INPUT_MIXER		0x20		/* bit mapped enables of functions */
										/* 0=enabled.  Bit 2=CD */
										/* Write 0x1B to enable CD */

/* SOUND BLASTER 16 DEFINITIONS */
/* Mixer Register Indexes */
#define SB16_MASTER_VOLUME_L	0x30		/* upper 5 bits are useful */
#define SB16_MASTER_VOLUME_R	0x31		/* upper 5 bits are useful */
#define SB16_VOICE_VOLUME_L		0x32		/* upper 5 bits are useful */
#define SB16_VOICE_VOLUME_R		0x33		/* upper 5 bits are useful */
#define SB16_MIDI_VOLUME_L		0x34		/* upper 5 bits are useful */
#define SB16_MIDI_VOLUME_R		0x35		/* upper 5 bits are useful */
#define SB16_CD_VOLUME_L		0x36		/* upper 5 bits are useful */
#define SB16_CD_VOLUME_R		0x37		/* upper 5 bits are useful */
#define SB16_LINE_VOLUME_L		0x38		/* upper 5 bits are useful */
#define SB16_LINE_VOLUME_R		0x39		/* upper 5 bits are useful */
#define SB16_MIC_VOLUME			0x3A		/* upper 5 bits are useful */
#define SB16_PC_SPEAKER_VOLUME	0x3B		/* upper 2 bits are useful */
#define SB16_OUTPUT_MIXER		0x3C		/* 7=reserved */
											/* 6=reserved */
											/* 5=reserved */
											/* 4=Line Left */
											/* 3=Line Right */
											/* 2=CD Left */
											/* 1=CD Right */
											/* 0=MIC */
#define SB16_INPUT_MIXER_LEFT	0x3D		/* 7=reserved */
											/* 6=MIDI Left */
											/* 5=MIDI Right */
											/* 4=Line Left */
											/* 3=Line Right */
											/* 2=CD Left */
											/* 1=CD Right */
											/* 0=MIC */
#define SB16_INPUT_MIXER_RIGHT	0x3E		/* 7=reserved */
											/* 6=MIDI Left */
											/* 5=MIDI Right */
											/* 4=Line Left */
											/* 3=Line Right */
											/* 2=CD Left */
											/* 1=CD Right */
											/* 0=MIC */
#define SB16_INPUT_GAIN_LEFT	0x3F		/* upper 2 bits are significant */
#define SB16_INPUT_GAIN_RIGHT	0x40		/* upper 2 bits are significant */
#define SB16_OUTPUT_GAIN_LEFT	0x41		/* upper 2 bits are significant */
#define SB16_OUTPUT_GAIN_RIGHT	0x42		/* upper 2 bits are significant */
#define SB16_AUTO_GAIN_CONTROL	0x43		/* 7-1=reserved */
											/* 0=AGC */
#define SB16_TREBLE_LEFT		0x44		/* upper 4 bits are significant */
#define SB16_TREBLE_RIGHT		0x45		/* upper 4 bits are significant */
#define SB16_BASS_LEFT			0x46		/* upper 4 bits are significant */
#define SB16_BASS_RIGHT			0x47		/* upper 4 bits are significant */

#define AUDIOBASE				0x220
#define DEFAULT_CD				"/dev/cd0"	/* default CD-ROM device */
#define CD_LEN					26			/* length of string for CD-ROM device */


/******** GLOBAL VARIABLES ********/
#ifdef DEFINE_GLOBALS
	#define GLOBAL
	#define INIT(x) = x
#else
	#define GLOBAL extern
	#define INIT(x)
#endif

GLOBAL fd_cd				INIT(-1);				/* file descriptor for cd-rom device */
GLOBAL BYTE devcd[CD_LEN]	INIT(DEFAULT_CD);		/* CD-ROM device */
GLOBAL pid_t timer_proxy	INIT(-1);				/* process id of proxy triggered by timer */
GLOBAL timer_t timer_id		INIT(-1);				/* identifier for real-time timer */
GLOBAL struct itimerspec playtime;					/* timer data for play time, stored during pause */
GLOBAL PtArg_t args[4];								/* data for Photon resources */

GLOBAL FLAG do_load			INIT(1);				/* 0 = CD loaded */
													/* 1 = load command required */
GLOBAL FLAG do_resume		INIT(0);				/* 1 = paused */
GLOBAL int playing_track	INIT(0);				/* track currently playing, 0=stopped */
GLOBAL int card_type		INIT(1);				/* type/model of audio card */
GLOBAL WORD mixer_base		INIT(AUDIOBASE + 4);	/* audio mixer port address */
GLOBAL FLAG do_audio_init;							/* 1 = audio card initialization required */
GLOBAL FLAG do_volume;								/* 1 = volume change required */
GLOBAL BYTE volume;									/* value for volume control */
GLOBAL FLAG do_treble;								/* 1 = treble command required */
GLOBAL BYTE treble;									/* value for bass control */
GLOBAL FLAG do_bass;								/* 1 = bass command required */
GLOBAL BYTE bass;									/* value for bass control */

#ifdef DEFINE_GLOBALS
	/* Map Slider Values To Mixer Values - Slider 0 is Mixer max */
	BYTE mixer_sb16[] = { 0xF8, 0xE0, 0xD0, 0xC0, 0xB0, 0xA0, 0x90, 0x80, 0x70, 0x60, 0x50, 0x40, 0x30, 0x20, 0x10, 0x00 };
	BYTE mixer_sbp[] = { 0xFF, 0xFF, 0xDD, 0xDD, 0xBB, 0xBB, 0x99, 0x99, 0x77, 0x77, 0x55, 0x55, 0x33, 0x33, 0x11, 0x11 };
	/* Valid I/O Port bases */
	WORD port_bases[] = { 0x224, 0x244, 0x264, 0x284, 0 };
	BYTE *cfg_file = "/usr/photon/config/playcd.%lu";
#else
	GLOBAL BYTE mixer_val[];
	GLOBAL WORD port_bases[];
	GLOBAL BYTE *cfg_file;
#endif
GLOBAL BYTE errbuf[81];								/* for error messages */
GLOBAL WORD new_mixer_base;							/* new mixer port value set in hardware dialog */
GLOBAL int new_card_type;							/* new type/model of audio card set in hardware dialog */
GLOBAL BYTE new_devcd[CD_LEN];						/* new CD-ROM device set in hardware dialog */

/* Audio Track Information */
GLOBAL struct cdrom_tochdr tochdr;					/* cd table of contents header data */
GLOBAL int must_be_zero 	INIT(55);
GLOBAL int total_tracks;							/* number of audio tracks */
GLOBAL struct cdrom_tocentry tkent[MAXTRKS];		/* info for each track */
GLOBAL struct cdrom_tocentry *tkentpt;
GLOBAL BYTE playlist[MAXTRKS + 1];					/* track numbers in order of selection, \0 is no more tracks */
GLOBAL BYTE trklistdata[MAXTRKS*12];				/* Audio Track Text For PtList Widget */
GLOBAL BYTE *trklist[MAXTRKS];						/* Pointers within trklistdata[] */
